#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
read and write local XML data in CMI format
"""

import os

from lxml import etree


def writer(root, file="cmif.xml", path="."):
    """
    write given root element to file at path
    """
    parser = etree.XMLParser(remove_blank_text=True)
    root.addprevious(etree.Comment(" Built with Python package cmif. "))
    xml = etree.ElementTree(root, parser=parser)
    if not os.path.exists(path):
        os.makedirs(path)
    out = os.path.join(path, file)
    xml.write(out, pretty_print=True, encoding="UTF-8", xml_declaration=True)


def reader(filepath):
    """
    read xml data from given file path
    """
    return etree.parse(filepath).getroot()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Copyright (c) 2020  Donatus Herre <git@herre.io>

This module offers the possibility to work with data in CMI format.
"""
__author__ = "Donatus Herre <git@herre.io>"
__version__ = "2020.6.25"
__license__ = "MIT"

from . import demo
from . import local
from . import build
from . import model
from . import retrieve
from . import authority
from . import service
from . import extract



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
extract XML data in CMI format
"""

import re
from .build import ns_cs, ns_xml


def title(data):
    """
    extract text of TEI element <title>
    """
    return data.find(".//title", namespaces=data.nsmap).text


def editor(data, multi=False):
    """
    | extract TEI element <editor>
    | set multi to True if multiple editors exist
    """
    return data.find(".//editor", namespaces=data.nsmap) if not multi else \
        data.findall(".//editor", namespaces=data.nsmap)


def editor_name(data, multi=False):
    """
    | extract text of TEI element <editor>
    | set multi to True if multiple editors exist
    """
    return editor(data, multi=multi).text.strip() if not multi else \
        [e.text.strip() for e in editor(data, multi=multi)]


def editor_email(data, multi=False):
    """
    | extract text of TEI element <email> from parent <editor>
    | set multi to True if multiple editors exist
    """
    return editor(data, multi=multi).find(".//email", namespaces=data.nsmap).text if not multi else \
        [e.find(".//email", namespaces=data.nsmap).text for e in editor(data, multi=multi)]


def publisher(data):
    """
    extract text from child <ref> of TEI element <publisher>
    """
    return data.find(".//publisher/ref", namespaces=data.nsmap).text


def publisher_target(data):
    """
    extract @target from child <ref> of TEI element <publisher>
    """
    return data.find(".//publisher/ref", namespaces=data.nsmap).attrib["target"]


def idno(data):
    """
    extract text of TEI element <idno>
    """
    return data.find(".//idno", namespaces=data.nsmap).text


def date(data):
    """
    extract @when from TEI element <date>
    """
    return data.find(".//date", namespaces=data.nsmap).attrib["when"]


def license(data):
    """
    extract text of TEI element <license>
    """
    return data.find(".//licence", namespaces=data.nsmap).text


def license_target(data):
    """
    extract @target from TEI element <licence>
    """
    return data.find(".//licence", namespaces=data.nsmap).attrib["target"]


def bibl(data, multi=False):
    """
    | extract TEI element <bibl>
    | set multi to True if multiple references exist
    """
    return data.find(".//bibl", namespaces=data.nsmap) if not multi else \
        data.findall(".//bibl", namespaces=data.nsmap)


def bibl_id(data, multi=False):
    """
    | extract @xml:id from TEI element <bibl>
    | set multi to True if multiple references exist
    """
    bibl_data = bibl(data, multi=multi)
    return bibl_data.attrib[ns_xml("id")] if not multi else \
        [b.attrib[ns_xml("id")] for b in bibl_data]


def bibl_type(data, multi=False):
    """
    | extract @type from TEI element <bibl>
    | set multi to True if multiple references exist
    """
    bibl_data = bibl(data, multi=multi)
    return bibl_data.attrib["type"] if not multi else \
        [b.attrib["type"] for b in bibl_data]


def bibl_text(data, multi=False):
    """
    | extract text of TEI element <bibl>
    | set multi to True if multiple references exist
    """
    bibl_data = bibl(data, multi=multi)
    return re.sub("[ \r\n]+", " ", "".join([l for l in list(bibl_data.itertext())]).strip()) if not multi else \
        [re.sub("[ \r\n]+", " ", "".join([l for l in list(b.itertext())]).strip()) for b in bibl_data]


def correspdesc(data):
    """
    extract TEI elements <correspDesc>
    """
    return data.findall(".//correspDesc", namespaces=data.nsmap)


def correspdesc_source(data):
    """
    extract @source from TEI elements <correspDesc>
    """
    correspdesc_data = correspdesc(data)
    try:
        return [cd.attrib["source"].replace("#", "") for cd in correspdesc_data]
    except KeyError:
        pass
    try:
        return [cd.attrib[ns_cs("source")].replace("#", "") for cd in correspdesc_data]
    except KeyError:
        pass
    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
search or retrieve GeoNames

you need to set the environment variable GEO_USER like this:

  $ export GEO_USER=[my-geonames-account-username]

or pass the variable username as parameter to the functions!
"""

import os
import requests

GEODATA = "http://api.geonames.org/getJSON"
GEOSEARCH = "http://api.geonames.org/searchJSON"


def request(geoname_id, style="full", lang=None, username=None):
    """request GeoNames entity by given id (requires GeoNames username)"""
    if username is None:
        try:
            username = os.environ["GEO_USER"]
        except KeyError:
            print("")
            print("to get the GeoNames JSON data either pass")
            print("your GeoNames username like this:")
            print("getJSON(\""+geoname_id+"\", username=USERNAME)\n")
            print("... or set environment variable GEO_USER")
            print("to your GeoNames username!")
            print("")
            return None
    params = {}
    params["geonameId"] = geoname_id
    params["style"] = style
    if lang is not None:
        params["lang"] = lang
    params["username"] = username
    try:
        result = requests.get(GEODATA, params=params).json()
    except requests.exceptions.ConnectionError:
        print("cmif.authority.geonames.request:")
        print("request failed! connection error...")
        return {}
    except requests.exceptions.Timeout:
        print("cmif.authority.geonames.request:")
        print("request failed! timeout... try later?")
        return {}
    except requests.exceptions.TooManyRedirects:
        print("cmif.authority.geonames.request:")
        print("request failed! bad url, try another one!")
        return {}
    except requests.exceptions.RequestException:
        print("cmif.authority.geonames.request:")
        print("request failed! don't know why...")
        return {}
    if "status" in result:
        print("failed to request GeoNames data!")
        print(result["status"]["message"])
        return {}
    return result


def search(place, bias="DE", username=None):
    """search GeoNames entity for given place (requires GeoNames username)"""
    if username is None:
        try:
            username = os.environ["GEO_USER"]
        except KeyError:
            print("")
            print("to do the GeoNames search either pass")
            print("your GeoNames username like this:")
            print("search(\""+place+"\", username=USERNAME)\n")
            print("... or set environment variable GEO_USER")
            print("to your GeoNames username!")
            print("")
            return None
    params = {}
    params["q"] = place
    params["username"] = username
    params["countryBias"] = bias
    params["style"] = "short"
    params["maxRows"] = 3
    try:
        result = requests.get(GEOSEARCH, params=params).json()["geonames"]
    except requests.exceptions.ConnectionError:
        print("cmif.authority.geonames.search:")
        print("request failed! connection error...")
        return "", ""
    except requests.exceptions.Timeout:
        print("cmif.authority.geonames.search:")
        print("request failed! timeout... try later?")
        return "", ""
    except requests.exceptions.TooManyRedirects:
        print("cmif.authority.geonames.search:")
        print("request failed! bad url, try another one!")
        return "", ""
    except requests.exceptions.RequestException:
        print("cmif.authority.geonames.search:")
        print("request failed! don't know why...")
        return "", ""
    if result == []:
        print("no results for given query!")
        return "", ""
    result = result[0]
    geoname = result["toponymName"]
    geonameId = "http://www.geonames.org/{0}".format(result["geonameId"])
    return geoname, geonameId



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
access data in CMI format through correspSearch webservices
"""

import requests
from lxml import etree

TEI_JSON = "http://correspSearch.bbaw.de/api/v1.1/tei-json.xql"
TEI_XML = "http://correspSearch.bbaw.de/api/v1.1/tei-xml.xql"

ENTITIES = {
  "HALLE": "http://www.geonames.org/2911522",
  "FORSTER": "http://d-nb.info/gnd/118534432",
  "HEMMERDE": "http://d-nb.info/gnd/1037496973"
}


def correspsearch(correspondent=None, sender=None, addressee=None,
                  startdate=None, enddate=None, place=None,
                  placeSender=None, placeAddressee=None,
                  cmiFile=None, publication=None,
                  available=None, strictDate=False, endpoint=TEI_XML):
    """
    | query data in CMI format via correspSearch API by given parameters
    | available endpoints: TEI_XML (default) / TEI_JSON
    | see https://correspsearch.net/index.xql?id=api for details
    """
    if correspondent == sender == addressee == publication == cmiFile \
            and sender is None:
        print("please specify one of the following parameters:")
        print("correspondent / sender / addressee / publication / cmiFile")
        return None
    params = {}
    if correspondent is not None:
        params["correspondent"] = correspondent
    if sender is not None:
        params["sender"] = sender
    if addressee is not None:
        params["addressee"] = addressee
    if publication is not None:
        params["publication"] = publication
        if correspondent is None:
            params["correspondent"] = "all"
    if cmiFile is not None:
        params["cmiFile"] = cmiFile
        if correspondent is None:
            params["correspondent"] = "all"
    if startdate is not None:
        params["startdate"] = startdate
    if enddate is not None:
        params["enddate"] = enddate
    if strictDate:
        params["strictDate"] = strictDate
    if place is not None:
        params["place"] = place
    if placeSender is not None:
        params["placeSender"] = placeSender
    if placeAddressee is not None:
        params["placeAddressee"] = placeAddressee
    if available is not None:
        params["available"] = available
    try:
        response = requests.get(endpoint, params=params)
        if response.status_code == 200:
            if endpoint == TEI_JSON:
                if type(response.json()) == dict:
                    return response.json()
                else:
                    print("Invalid query!")
                    return {}
            else:
                tei = etree.fromstring(response.content)
                if tei.text != "Invalid query.":
                    return tei
                else:
                    print("cmif.service.correspsearch:")
                    print("Invalid query!")
                    return None
        else:
            print("cmif.service.correspsearch:")
            print("request is not ok!")
            print("url of request:")
            print(response.url)
            print("http status code: ",
                  str(response.status_code))
    except requests.exceptions.ConnectionError:
        print("cmif.service.correspsearch:")
        print("request failed! connection error...")
    except requests.exceptions.Timeout:
        print("cmif.service.correspsearch:")
        print("request failed! timeout... try later?")
    except requests.exceptions.TooManyRedirects:
        print("cmif.service.correspsearch:")
        print("request failed! bad url, try another one!")
    except requests.exceptions.RequestException:
        print("cmif.service.correspsearch:")
        print("request failed! don't know why...")
    return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
schema models of CMI format
"""

from .local import reader

import os

from lxml.etree import RelaxNG


def odd():
    """
    get ODD specification of CMI format
    """
    return reader(os.path.join(os.path.dirname(__file__),
                               "standard/odd/cmi-customization.xml"))


def rng():
    """
    get RNG schema of CMI format
    """
    return RelaxNG(
            reader(os.path.join(os.path.dirname(__file__),
                                "standard/schema/cmi-customization.rng")))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
build XML data in CMI format
"""

import uuid
from lxml import etree

XML_NS = "http://www.w3.org/XML/1998/namespace"
TEI_NS = "http://www.tei-c.org/ns/1.0"
CS_NS = "http://www.bbaw.de/telota/correspSearch"
RNG_SCHEMA = "https://raw.githubusercontent.com/TEI-Correspondence-SIG/" + \
    "CMIF/master/schema/cmi-customization.rng"
PI_TEXT = "href=\""+RNG_SCHEMA+"\" type=\"application/xml\" schematypens" + \
    "=\"http://relaxng.org/ns/structure/1.0\""
CC_TEXT = "This file is licensed under the terms " \
            + "of the Creative-Commons-License CC-BY 4.0"
CC_URL = "https://creativecommons.org/licenses/by/4.0/"


def pi_rng():
    """
    create processing instruction <?xml-model?>
    """
    return etree.ProcessingInstruction("xml-model", PI_TEXT)


def ns_xml(attrib):
    """
    add xml namespace to given attribute
    """
    return "{" + XML_NS + "}" + attrib


def ns_cs(attrib):
    """
    add correspSearch namespace to given attribute
    """
    return "{" + CS_NS + "}" + attrib


def tei_root(children=None):
    """
    create TEI root element <TEI> with (optional) children
    """
    root = etree.Element("TEI")
    root.set("xmlns", TEI_NS)
    add_children(root, children)
    return root


def tei_header(children=None):
    """
    create TEI element <teiHeader> with (optional) children
    """
    header = etree.Element("teiHeader")
    add_children(header, children)
    return header


def tei_file_desc(children=None):
    """
    create TEI element <fileDesc> with (optional) children
    """
    file_desc = etree.Element("fileDesc")
    add_children(file_desc, children)
    return file_desc


def tei_title_stmt(children=None):
    """
    create TEI element <titleStmt> with (optional) children
    """
    title_stmt = etree.Element("titleStmt")
    add_children(title_stmt, children)
    return title_stmt


def tei_title(elem_text):
    """
    crate TEI element <title> with given element text
    """
    title = etree.Element("title")
    title.text = elem_text
    return title


def tei_editor(elem_text):
    """
    create TEI element <editor> with given element text
    """
    editor = etree.Element("editor")
    editor.text = elem_text
    return editor


def tei_email(elem_text):
    """
    create TEI element <email> with given element text
    """
    email = etree.Element("email")
    email.text = elem_text
    return email


def tei_publication_stmt(children=None):
    """
    create TEI element <publicationStmt> with (optional) children
    """
    publication_stmt = etree.Element("publicationStmt")
    add_children(publication_stmt, children)
    return publication_stmt


def tei_publisher(child_ref=None):
    """
    create TEI element <publisher> with (optional) child element <ref>
    """
    publisher = etree.Element("publisher")
    add_child(publisher, child_ref)
    return publisher


def tei_ref(elem_text, attrib_target):
    """
    create TEI element <ref> with @target
    """
    ref = etree.Element("ref")
    ref.set("target", attrib_target)
    ref.text = elem_text
    return ref


def tei_idno(elem_text, attrib_type="url"):
    """
    create TEI element <idno> with @type
    """
    idno = etree.Element("idno")
    idno.set('type', attrib_type)
    idno.text = elem_text
    return idno


def tei_availability(child_license=None):
    """
    create TEI element <availability> with (optional) child element <licence>
    """
    availability = etree.Element("availability")
    add_child(availability, child_license)
    return availability


def tei_license(elem_text="", attrib_target=""):
    """
    create TEI element <licence> with (optional) text and @target
    """
    if elem_text == "" and attrib_target == "":
        elem_text = CC_TEXT
        attrib_target = CC_URL
    license = etree.Element("licence")
    license.set("target", attrib_target)
    license.text = elem_text
    return license


def tei_source_desc(children=None):
    """
    create TEI element <sourceDesc> with (optional) children
    """
    source_desc = etree.Element("sourceDesc")
    add_children(source_desc, children)
    return source_desc


def tei_bibl(elem_text, attrib_type, attrib_xml_id=None, domain=None):
    """
    create TEI element <bibl> with given text and @type
    """
    bibl = etree.Element("bibl")
    if attrib_type not in ["print", "online", "hybrid"]:
        print("@type has to be 'print', 'online' or 'hybrid'!")
        return None
    bibl.set("type", attrib_type)
    bibl.text = elem_text
    if attrib_xml_id is None:
        attrib_xml_id = tei_bibl_id(domain=domain)
    bibl.set(ns_xml("id"), attrib_xml_id)
    return bibl


def tei_bibl_id(domain=None):
    """
    | generate uuid for @xml:id of TEI element <bibl> by given domain
    | if domain is None a random uuid is being generated
    """
    found = False
    if domain is not None:
        while not found:
            domain_uuid = str(uuid.uuid3(uuid.NAMESPACE_URL, domain))
            if domain_uuid[0].isalpha():
                return domain_uuid
            domain_uuid = str(uuid.uuid3(uuid.NAMESPACE_DNS, domain))
            if domain_uuid[0].isalpha():
                return domain_uuid
            domain_uuid = str(uuid.uuid3(uuid.NAMESPACE_OID, domain))
            if domain_uuid[0].isalpha():
                return domain_uuid
            domain_uuid = str(uuid.uuid3(uuid.NAMESPACE_X500, domain))
            if domain_uuid[0].isalpha():
                return domain_uuid
            domain_uuid = str(uuid.uuid5(uuid.NAMESPACE_URL, domain))
            if domain_uuid[0].isalpha():
                return domain_uuid
            domain_uuid = str(uuid.uuid5(uuid.NAMESPACE_DNS, domain))
            if domain_uuid[0].isalpha():
                return domain_uuid
            domain_uuid = str(uuid.uuid5(uuid.NAMESPACE_OID, domain))
            if domain_uuid[0].isalpha():
                return domain_uuid
            domain_uuid = str(uuid.uuid5(uuid.NAMESPACE_X500, domain))
            if domain_uuid[0].isalpha():
                return domain_uuid
            if not found:
                print("could not create uuid valid as xml:id from given domain!")
                print("try another one? meanwhile generating a random uuid...")
            while not found:
                domain_uuid = str(uuid.uuid4())
                if domain_uuid[0].isalpha():
                    return domain_uuid
    else:
        print("no domain given! generating a random uuid...")
        while not found:
            domain_uuid = str(uuid.uuid4())
            if domain_uuid[0].isalpha():
                return domain_uuid


def tei_profile_desc(children=None):
    """
    create TEI element <profileDesc>
    """
    profile_desc = etree.Element("profileDesc")
    add_children(profile_desc, children)
    return profile_desc


def tei_corresp_desc(attrib_key="", attrib_ref="",
                     attrib_source="", children=None):
    """
    create TEI element <correspDesc> with @ref
    """
    corresp_desc = etree.Element("correspDesc")
    add_attrib(corresp_desc, "key", attrib_key)
    add_attrib(corresp_desc, "ref", attrib_ref)
    add_attrib(corresp_desc, "source", attrib_source)
    add_children(corresp_desc, children)
    return corresp_desc


def tei_corresp_action(attrib_type, children=None):
    """
    create TEI element <correspAction> with @type
    """
    corresp_action = etree.Element("correspAction")
    if attrib_type not in ["sent", "received"]:
        print("@type has to be 'sent' or 'received'!")
        return None
    corresp_action.set("type", attrib_type)
    add_children(corresp_action, children)
    return corresp_action


def tei_date(attrib_when="", attrib_from="", attrib_to="",
             attrib_not_before="", attrib_not_after=""):
    """
    create TEI element <date> with @when or @from and @to
    """
    date = etree.Element("date")
    add_attrib(date, "when", attrib_when)
    add_attrib(date, "from", attrib_from)
    add_attrib(date, "to", attrib_to)
    add_attrib(date, "notBefore", attrib_not_before)
    add_attrib(date, "notAfter", attrib_not_after)
    return date


def tei_place_name(elem_text, attrib_ref=""):
    """
    create TEI element <placeName> with given element text and @ref
    """
    place_name = etree.Element("placeName")
    place_name.text = elem_text
    add_attrib(place_name, "ref", attrib_ref)
    return place_name


def tei_pers_name(elem_text, attrib_ref=""):
    """
    create TEI element <persName> with given element text and @ref
    """
    pers_name = etree.Element("persName")
    pers_name.text = elem_text
    add_attrib(pers_name, "ref", attrib_ref)
    return pers_name


def tei_org_name(elem_text, attrib_ref=""):
    """
    create TEI element <orgName> with given element text
    """
    org_name = etree.Element("orgName")
    org_name.text = elem_text
    add_attrib(org_name, "ref", attrib_ref)
    return org_name


def tei_text_empty():
    """
    create TEI element <text> with child elements <body> and <p> (empty)
    """
    text = tei_text()
    body = tei_body()
    body.append(tei_p())
    text.append(body)
    return text


def tei_text():
    """
    create TEI element <text>
    """
    return etree.Element("text")


def tei_body():
    """
    create TEI element <body>
    """
    return etree.Element("body")


def tei_p():
    """
    create TEI element <p>
    """
    return etree.Element("p")


def add_pi(tree):
    """
    add processing instruction <?xml-model?> to given element tree
    """
    tree.getroot().addprevious(pi_rng())


def add_attrib(element, name, value):
    """
    add attribute to element if value != ""
    """
    if value != "":
        element.set(name, value)


def add_child(parent, element):
    """
    add element to parent if not None
    """
    if element is not None:
        parent.append(element)


def add_children(parent, elements):
    """
    add elements to parent if not None
    """
    if elements is not None:
        for child in elements:
            parent.append(child)


def pretty(elements):
    """
    pretty print given elements
    """
    print(etree.tostring(elements, pretty_print=True).decode().strip())


def tostr(element):
    """
    convert given element to str
    """
    return etree.tostring(element).decode().strip()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
demo XML data in CMI format
"""

from .retrieve import remote_file

WIELAND = "https://correspsearch.net/storage/WielandBW-1.xml"
GOTTSCHED = "https://raw.githubusercontent.com/" + \
    "saw-leipzig/cmif-gottsched/master/letters.xml"


def wieland():
    """
    get data of Wieland's Briefe (1750-1760) as demo for CMI-format
    """
    return remote_file(WIELAND)


def gottsched():
    """
    get data of Gottsched's Briefwechsel as demo for CMI-format
    """
    return remote_file(GOTTSCHED)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
retrieve remote XML data in CMI format
"""

import requests

from lxml import etree


def remote_file(url):
    """
    send http get request to given url
    """
    try:
        response = requests.get(url)
        if response.status_code == 200:
            return etree.fromstring(response.content)
        print("requesting remote file failed!")
        print("url of request:")
        print(response.url)
        print("http status code: ",
              str(response.status_code))
    except requests.exceptions.ConnectionError:
        print("request failed! connection error...")
    except requests.exceptions.Timeout:
        print("request failed! timeout... try later?")
    except requests.exceptions.TooManyRedirects:
        print("request failed! bad url, try another one!")
    except requests.exceptions.RequestException:
        print("request failed! don't know why...")
    return None
